/*
 * poppassd.c
 *
 * A Eudora and NUPOP change password server for Digital Unix with C2 security
 * Ported from Solaris version by Peter Mayne, the man from DEC.
 *
 * Modified by Tarjei T. Jensen
 *
 * Tested with Digital Unix 3.2C and Qualcomm's Eudora
 *
 * Some of the original Solaris comments:
 *
 * Graziano Sommariva
 * Ansaldo Sistemi Informatici Genova Italy.
 * sommariva@ansaldo.it
 *
 * Based on earlier versions by 
 * John Norstad
 * j-norstad@nwu.edu
 *
 *
 * Note that unencrypted passwords are transmitted over the network.  If
 * this bothers you, think hard about whether you want to implement the
 * password changing feature.  On the other hand, it's no worse than what
 * happens when you run /bin/passwd while connected via telnet or rlogin.
 * Well, maybe it is, since the use of a dedicated port makes it slightly
 * easier for a network snooper to snarf passwords off the wire.
 *
 * NOTE: In addition to the security issue outlined in the above paragraph,
 * you should be aware that this program is going to be run as root by
 * ordinary users and it mucks around with the password file.  This should
 * set alarms off in your head.  I think I've devised a pretty foolproof
 * way to ensure that security is maintained, but I'm no security expert and
 * you would be a fool to install this without first reading the code and
 * ensuring yourself that what I consider safe is good enough for you.  If
 * something goes wrong, it's your fault, not mine.
 *
 *        
 * Should be owned by root, and executable only by root.  It can be started
 * with an entry in /etc/inetd.conf such as the following:
 *
 * poppassd stream tcp nowait root /usr/local/bin/poppassd poppassd
 * 
 * and in /etc/services:
 * 
 * poppassd	106/tcp
 *
 * Logs to the local2 facility. Should have an entry in /etc/syslog.conf
 * like the following:
 *
 * local2.err	/var/adm/poppassd-log
 *
 *	Changed #include <strings.h> to <string.h> and moved varargs.h
 *	above the syslog.h include.
 *
 *	I also added a timeout during input.  [mark@qualcomm.com]
 *
 */
 
 
/* Steve Dorner's description of the simple protocol:
 *
 * The server's responses should be like an FTP server's responses; 
 * 1xx for in progress, 2xx for success, 3xx for more information
 * needed, 4xx for temporary failure, and 5xx for permanent failure.  
 * Putting it all together, here's a sample conversation:
 *
 *   S: 200 hello\r\n
 *   E: user yourloginname\r\n
 *   S: 300 please send your password now\r\n
 *   E: pass yourcurrentpassword\r\n
 *   S: 200 My, that was tasty\r\n
 *   E: newpass yournewpassword\r\n
 *   S: 200 Happy to oblige\r\n
 *   E: quit\r\n
 *   S: 200 Bye-bye\r\n
 *   S: <closes connection>
 *   E: <closes connection>
 */
 
#define VERSION "2.0.a"

#define BUFSIZE 512

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/security.h>
#include <prot.h>

static void WriteToClient (char *fmt, ...);
static void ReadFromClient(char *line);
static int chkPass(struct pr_passwd *pr, char *p);
static void getSalt(char *salt);

main (int argc, char *argv[])
{
char line[BUFSIZE];
char user[BUFSIZE];
char oldpass[BUFSIZE];
char newpass[BUFSIZE];
struct pr_passwd *pr_pwd, pr_newpwd;
/*
int c;
*/
char salt[3];
     
	set_auth_parameters (argc, argv);

	*user = *oldpass = *newpass = 0;
     
	if (openlog ("poppassd", LOG_PID, LOG_LOCAL2) < 0)
	{
		WriteToClient ("500 Can't open syslog.");
		exit (1);
	}
     
	WriteToClient ("200 poppassd v%s for Digital Unix with C2 security Hello, who are you?",
                       VERSION);
	ReadFromClient (line);
	sscanf (line, "user %s", user) ;
	if (strlen (user) == 0)
	{
		WriteToClient ("500 Username required.");
		exit(1);
	}

	WriteToClient ("200 your password please.");
	ReadFromClient (line);
	sscanf (line, "pass %s", oldpass) ;
	if (strlen (oldpass) == 0)
	{
		WriteToClient ("500 Password required.");
		exit(1);
	}
     
	if ((pr_pwd = getprpwnam (user)) == NULL)
	{
		WriteToClient ("500 Unknown user, %s.", user);
		exit(1);
	}

	if (!chkPass (pr_pwd, oldpass))
	{
		WriteToClient ("500 Old password is incorrect.");
		exit(1);
	}

	WriteToClient ("200 your new password please.");
	ReadFromClient (line);
	sscanf (line, "newpass %s", newpass);
     
	if (strlen (newpass) == 0)
	{
		WriteToClient ("500 New password required.");
		exit(1);
	}

	if (strcmp (oldpass, newpass) == 0)
	{
		WriteToClient ("500 New password must be different.");
		exit (1);
	}

	pr_newpwd = *pr_pwd;
	pr_newpwd.ufld.fd_schange = time (NULL);
	pr_newpwd.uflg.fg_schange = 1;
	pr_newpwd.ufld.fd_pwchanger = pr_newpwd.ufld.fd_uid;
	pr_newpwd.uflg.fg_pwchanger = 1;
	getSalt (salt);
	strcpy (pr_newpwd.ufld.fd_encrypt, bigcrypt (newpass, salt));
	if (putprpwnam (user, &pr_newpwd))
	{
		syslog (LOG_ERR, "password changed for %s", user);
		WriteToClient ("200 Password changed, thank you.");

		ReadFromClient (line);
		if (strncmp(line, "quit", 4) != 0)
		{
			WriteToClient("500 Quit required.");
			exit (1);
		}
	  
		WriteToClient("200 Bye.");
		exit (0);
	}
	else
	{
		syslog (LOG_ERR, "failed attempt by %s", user);
		WriteToClient ("500 Unable to change password");
		exit(1);
	}
}

static void WriteToClient (char *fmt, ...)
{
va_list args;
char *str;
	
	va_start (args, fmt);
	vfprintf (stdout, fmt, args);
	va_end (args);
	fputs ("\r\n", stdout);
	fflush (stdout);
}

static void ReadFromClient(char *line)
{
char *sp;
int i;

	strcpy (line, "");
	fgets (line, BUFSIZE, stdin);
	if ((sp = strchr(line, '\n')) != NULL) *sp = '\0'; 
	if ((sp = strchr(line, '\r')) != NULL) *sp = '\0'; 
	
	/* convert initial keyword on line to lower case. */
	
	for (sp = line; isupper(*sp); sp++)
		*sp = tolower(*sp);
}

static int chkPass(struct pr_passwd *pr, char *p)
{
    /*
	We don't accept connections from users with null passwords.
	Compare the supplied password with the password file entry.
	*/
    if ((pr->ufld.fd_encrypt == NULL) || (pr->ufld.fd_encrypt[0] == '\0') ||
		strcmp (bigcrypt (p, pr->ufld.fd_encrypt),
pr->ufld.fd_encrypt))
    {
        return (0);
    }

    return(1);
}

static void getSalt(char *salt)
{
static char *saltset=
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890./";
struct timespec tp;

	getclock (TIMEOFDAY, &tp);
	srandom ((unsigned)(tp.tv_sec-tp.tv_nsec));

	salt[0] = saltset[random () % 64];
	salt[1] = saltset[random () % 64];
	salt[2] = '\0';
}
